"""This module contains utilities for generating Python code from AsyncAPI specifications."""

from __future__ import annotations

from ast import Assign, Attribute, Call, Constant, ImportFrom, Load, Name, Store, alias

from zen_generator.generators.common_python import BasePythonGenerator


class Generator:
    """Class for generating Python code from AsyncAPI specification.

    Attributes:
        asyncapi_file (Path): The path to the AsyncAPI file.
        output_file (Path): The path to the generated Python file.
        models_file (Path): The path to the generated models file.
        override_base_class (str | None): The base class to override in the generated code.
        decorator_list (Sequence[expr]): A list of decorators to apply to the generated class.
        additional_imports (Sequence[stmt | ImportFrom]): Additional imports to include in the generated code.
        extra_assignments (Sequence[stmt]): Additional assignments to include in the generated code.

    Methods:
        generate_files_from_asyncapi: Generate Python files from an AsyncAPI specification.
    """

    @staticmethod
    def fastapi_generator() -> BasePythonGenerator:
        """Generate a FastAPI generator from an AsyncAPI specification.

        Returns:
            BasePythonGenerator: A generator for generating FastAPI code from an AsyncAPI specification.
        """
        return BasePythonGenerator(
            override_base_class="BaseModel",
            extra_imports=[
                ImportFrom(module="fastapi", names=[alias(name="FastAPI")], level=0),
            ],
            extra_assignments=[
                Assign(
                    targets=[Name(id="app", ctx=Store())],
                    value=Call(func=Name(id="FastAPI", ctx=Load()), args=[], keywords=[]),
                )
            ],
            decorator_list=[
                Call(
                    func=Attribute(value=Name(id="app", ctx=Load()), attr="get", ctx=Load()),
                    args=[Constant(value="/{func_name}")],
                    keywords=[],
                )
            ],
        )

    @staticmethod
    def pure_python_generator() -> BasePythonGenerator:
        """Generate a pure Python generator from an AsyncAPI specification.

        Returns:
            BasePythonGenerator: A generator for generating pure Python code from an AsyncAPI specification.
        """
        return BasePythonGenerator(
            extra_imports=[
                ImportFrom(module="typing", names=[alias(name="TypedDict")], level=0),
                ImportFrom(module="utils.enums", names=[alias(name="Choices")], level=0),
            ]
        )



"""This module contains utilities for parsing Python code.

The functions in this module provide a higher-level interface than the `ast` module,
and are used to generate Python code from AsyncAPI specifications.

"""

from __future__ import annotations

import re
from ast import AsyncFunctionDef, FunctionDef, Module, get_docstring, walk
from typing import Any

from zen_generator.core.ast_utils import convert_annotations_to_asyncapi_schemas, parse_type_annotation

# Define regular expressions for Args and Returns sections
ARGS_PATTERN = re.compile(r"Args:(.*?)(?:Returns|$)", re.DOTALL)
RETURNS_PATTERN = re.compile(r"Returns:(.*?)$", re.DOTALL)
ARGS_DESCRIPTION_PATTERN = r"\s*([\w_]+)\s*\(\)\s*:\s*([^(\n)]+)"


def function_to_asyncapi_schemas(
    function_node: FunctionDef | AsyncFunctionDef,
) -> tuple[dict[str, Any], dict[str, Any]]:
    """Convert a function definition to AsyncAPI request and response schemas.

    Args:
        function_node: The function definition node

    Returns:
        A tuple containing the request and response schemas for the function
    """
    name = function_node.name
    docstring = get_docstring(function_node) or ""

    args_match = ARGS_PATTERN.search(docstring)
    returns_match = RETURNS_PATTERN.search(docstring)

    parameters_description = {}
    if args_match:
        args_section = args_match.group(1).strip()
        parameters_description = dict(re.findall(ARGS_DESCRIPTION_PATTERN, args_section))

    request_schema = _create_request_schema(function_node, name, docstring, parameters_description)
    response_schema = _create_response_schema(function_node, name, docstring, returns_match)

    return request_schema, response_schema


def _create_request_schema(
    function_node: FunctionDef | AsyncFunctionDef,
    name: str,
    docstring: str,
    parameters_description: dict,
) -> dict[str, Any]:
    """Convert a function definition to AsyncAPI request schema.

    Args:
        function_node: The function definition node
        name: The name of the function
        docstring: The docstring of the function
        parameters_description: A dictionary of parameter descriptions

    Returns:
        A dictionary representing the request schema
    """
    request_payload: dict[str, Any] = {
        "type": "object",
        "required": [],
        "properties": {},
    }

    request_schema = {
        "title": f"Request params for {name}",
        "summary": "",
        "description": docstring,
        "payload": request_payload,
    }

    for param in function_node.args.args:
        property_name = param.arg
        items = parse_type_annotation(param.annotation)
        conv = convert_annotations_to_asyncapi_schemas(items)
        description = parameters_description.get(param.arg, "")

        if conv.get("required"):
            request_payload["required"].append(property_name)
            del conv["required"]

        request_payload["properties"][property_name] = conv.get("properties")
        request_payload["properties"][property_name]["description"] = description

    return request_schema


def _create_response_schema(
    function_node: FunctionDef | AsyncFunctionDef,
    name: str,
    docstring: str,
    returns_match: re.Match | None,
) -> dict[str, Any]:
    """Convert a function definition to AsyncAPI response schema.

    Args:
        function_node: The function definition node
        name: The name of the function
        docstring: The docstring of the function
        returns_match: A regular expression match object for the "Returns" section

    Returns:
        A dictionary representing the response schema
    """
    response_description = ""
    if returns_match:
        response_description = returns_match.group(1).strip()
    elif "Returns:" in docstring:
        start_returns = docstring.find("Returns:") + len("Returns:")
        response_description = docstring[start_returns:].strip()

    response_schema = {
        "title": f"Response params for {name}",
        "summary": "",
        "description": response_description,
    }

    return_type = parse_type_annotation(function_node.returns)
    properties = convert_annotations_to_asyncapi_schemas(return_type)
    response_payload = properties.get("properties", {})
    response_schema["payload"] = response_payload

    if properties.get("required") and isinstance(response_payload, dict):
        response_payload["format"] = "required"

    return response_schema


def function_content_reader(tree: Module | None) -> tuple[str | None, dict[str, Any]]:
    """Read function content from a tree of nodes.

    Args:
        tree: The root node of the tree

    Returns:
        A tuple containing the docstring of the tree and a dictionary of functions
        where the keys are the names of the functions and the values are dictionaries
        containing the request and response schemas
    """
    functions_docstring = get_docstring(tree) if tree else ""
    functions_to_async: dict[str, Any] = {}

    if tree is not None:
        for node in walk(tree):
            if isinstance(node, (FunctionDef, AsyncFunctionDef)):
                request, response = function_to_asyncapi_schemas(node)
                functions_to_async.update({f"{node.name}": {"request": request, "response": response}})

    return functions_docstring, functions_to_async



from __future__ import annotations

ASYNCAPI_TO_PYTHON_TYPES: dict[str, str] = {
    "string": "str",
    "integer": "int",
    "boolean": "bool",
    "array": "list",
    "object": "object",
}


def convert_asyncapi_to_python(asyncapi_type: str) -> str:
    """Convert a AsyncAPI type to the corresponding Python type.

    Args:
        asyncapi_type (str): The AsyncAPI type

    Returns:
        str: The corresponding Python type
    """
    return ASYNCAPI_TO_PYTHON_TYPES.get(asyncapi_type, asyncapi_type)


def convert_python_to_asyncapi(python_type: str) -> str | None:
    """Convert a Python type to the corresponding AsyncAPI type.

    Args:
        python_type (str): The Python type

    Returns:
        str | None: The corresponding AsyncAPI type
    """
    reverse_mapping = {py: asyncapi for asyncapi, py in ASYNCAPI_TO_PYTHON_TYPES.items()}
    return reverse_mapping.get(python_type)



from __future__ import annotations

from pathlib import Path

import typer
from rich import print
from typing_extensions import Annotated

from zen_generator.generators.asyncapi import generate_asyncapi_from_files
from zen_generator.generators.python import Generator

app = typer.Typer()


@app.command()
def asyncapi_documentation(
    models_file: Annotated[Path, typer.Option()] = Path("models.py"),
    functions_file: Annotated[Path, typer.Option()] = Path("functions.py"),
    output_file: Annotated[Path, typer.Option()] = Path("asyncapi.yaml"),
    application_name: Annotated[str, typer.Option()] = "Zen",
) -> None:
    """Generate AsyncAPI documentation from source code.

    Generate the AsyncAPI documentation from the source code in the models_file
    and functions_file. The output will be a yaml file saved in the output_file
    with the name of the application as title.

    Args:
        models_file: The path to the file containing the models.
        functions_file: The path to the file containing the functions.
        output_file: The path to the output file.
        application_name: The name of the application.
    """
    print("Preparing to generate the documentation")
    if models_file.is_file() and functions_file.is_file():
        generate_asyncapi_from_files(models_file, functions_file, output_file, application_name)
    else:
        print(
            f":boom: :boom: [bold red]the source file '{models_file}' "
            f"or the file '{functions_file}' is not a file![/bold red]"
        )
        raise typer.Abort()


@app.command()
def pure_python(
    asyncapi_file: Annotated[Path, typer.Option()] = Path("asyncapi.yaml"),
    models_file: Annotated[Path, typer.Option()] = Path("models.py"),
    functions_file: Annotated[Path, typer.Option()] = Path("functions.py"),
    application_name: Annotated[str, typer.Option()] = "Zen",
    is_async: Annotated[bool, typer.Option()] = False,
) -> None:
    """Generate pure Python models and functions from AsyncAPI file.

    Generate the models and functions from the AsyncAPI file in the asyncapi_file.
    The output will be two files, one for the models and one for the functions,
    saved in the models_file and functions_file respectively.

    Args:
        asyncapi_file: The path to the AsyncAPI file.
        models_file: The path to the output file for the models.
        functions_file: The path to the output file for the functions.
        application_name: The name of the application.
        is_async: Whether the generated functions should be async or not.
    """
    print("Preparing to generate models and functions from the asyncapi file")
    if asyncapi_file.is_file():
        generator = Generator.pure_python_generator()
        generator.generate_files_from_asyncapi(asyncapi_file, models_file, functions_file, application_name, is_async)
    else:
        print(
            f":boom: :boom: [bold red]the source file '{asyncapi_file}' "
            f"or the file '{functions_file}' is not a file![/bold red]"
        )
        raise typer.Abort()


@app.command()
def fastapi(
    asyncapi_file: Annotated[Path, typer.Option()] = Path("asyncapi.yaml"),
    models_file: Annotated[Path, typer.Option()] = Path("models.py"),
    functions_file: Annotated[Path, typer.Option()] = Path("functions.py"),
    application_name: Annotated[str, typer.Option()] = "Zen",
    is_async: Annotated[bool, typer.Option()] = False,
) -> None:
    """Generate FastAPI models and functions from AsyncAPI file.

    Generate the models and functions from the AsyncAPI file in the asyncapi_file.
    The output will be two files, one for the models and one for the functions,
    saved in the models_file and functions_file respectively.

    Args:
        asyncapi_file: The path to the AsyncAPI file.
        models_file: The path to the output file for the models.
        functions_file: The path to the output file for the functions.
        application_name: The name of the application.
        is_async: Whether the generated functions should be async or not.

    """
    print("Preparing to generate models and functions from the asyncapi file")
    if asyncapi_file.is_file():
        generator = Generator.fastapi_generator()
        generator.generate_files_from_asyncapi(asyncapi_file, models_file, functions_file, application_name, is_async)
    else:
        print(
            f":boom: :boom: [bold red]the source file '{asyncapi_file}' "
            f"or the file '{functions_file}' is not a file![/bold red]"
        )
        raise typer.Abort()


@app.callback()
def main() -> None:
    print("Welcome to the zen generator")
    app()


if __name__ == "__main__":
    app()
